/********************************************************************
 *
 * Copyright (C) 2014 Microchip Technology Inc. and its subsidiaries
 * (Microchip).  All rights reserved.
 *
 * You are permitted to use the software and its derivatives with Microchip
 * products. See the license agreement accompanying this software, if any, for
 * more info about your rights and obligations.
 *
 * SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR
 * PURPOSE. IN NO EVENT SHALL MICROCHIP, SMSC, OR ITS LICENSORS BE LIABLE OR
 * OBLIGATED UNDER CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH
 * OF WARRANTY, OR OTHER LEGAL EQUITABLE THEORY FOR ANY DIRECT OR INDIRECT
 * DAMAGES OR EXPENSES INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES, OR OTHER SIMILAR COSTS. To the fullest
 * extend allowed by law, Microchip and its licensors liability will not exceed
 * the amount of fees, if any, that you paid directly to Microchip to use this
 * software.
 *************************************************************************
 *
 *                           dali.c
 *
 * About:
 *  DALI protocol implementation. This provides the machinery to send/receive
 *  data on the DALI bus acting mostly as a Control Device (master). However
 *  it can also properly receive data sent by other Control Devices and give
 *  it to the application, along with any backframe the other Control Device may
 *  have received from existing Control Gears (slaves). When configured to do
 *  so it can send arbitrary digital waveforms on the DALI bus and can listen
 *  for DALI backframes that may come. It can also capture data on the bus as a
 *  digital waveform and forward it to the application.
 *
 * Hardware:
 *  USB to DALI Interface - DM160215
 *
 *
 * Author            Date            Ver     Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Mihai Cuciuc      18 October 2013  1.00    Implemented DALI protocol
 *
 *************************************************************************/


#include "dali.h"

// Timer flags and registers
#define TICK_TIMER_IF                   TMR2IF                          // 1ms timer tick flag
#define TE_TIMER_IF                     TMR1IF                          // Te timer flag
#define TE_TIMER_ON                     TMR1ON                          // Te timer enable bit
#define TE_TIMER_VAL                    TMR1                            // Te timer counting register

// Interrupt edge configuration and query. The values for the rising/falling
// edges are inverted since there's a layer of inversion in the optocouplers'
// transistors on the board. Thus, what we in the software call "rising edge"
// is a falling edge on the PIC's pin, but a rising edge on the DALI bus.
#define INT_DALI_IS_FALLING()           (OPTION_REGbits.INTEDG == 1)    // Check whether external interrupt is configured to trigger on the DALI falling edge
#define INT_DALI_IS_RISING()            (OPTION_REGbits.INTEDG == 0)    // Check whether external interrupt is configured to trigger on the DALI rising edge
#define INT_DALI_FALLING()              OPTION_REGbits.INTEDG = 1       // Configure external interrupt to trigger on the DALI falling edge
#define INT_DALI_RISING()               OPTION_REGbits.INTEDG = 0       // Configure external interrupt to trigger on the DALI rising edge
#define INT_DALI_TOGGLE()               OPTION_REGbits.INTEDG ^= 1      // Toggle external interrupt edge
#define INT_IF                          INTF                            // External interrupt flag


#define DALI_RX                         RC1                             // RX line input register bit
#define DALI_RX_TRIS                    TRISC1                          // RX line tristate register bit
#define DALI_TX                         LATA5                           // TX line output register bit
#define DALI_TX_TRIS                    TRISA5                          // TX line tristate register bit

// There's a layer of inversion in the optocoupelrs' transistors on the board
// and we define these two symbols such that other hardware configurations can
// easily be adopted
#define DALI_HI                         0
#define DALI_LO                         1

// The number of timer clocks that make up a 416us delay. The timer is configured
// to run at 3MHz, thus our delay will be 3MHz * 416.(6)us = 1250
#define TE                              1250
#define TE_TIMER_OVERFLOW_VAL           65536

// Various timer values that need to be loaded into the timer register to produce
// a given delay until the timer overflows.
#define TE_VALUE                        (TE_TIMER_OVERFLOW_VAL - TE)
#define TE4_VALUE                       (TE_TIMER_OVERFLOW_VAL - 4 * TE)
#define TE22_VALUE                      (TE_TIMER_OVERFLOW_VAL - 22 * TE)
#define TE_ERR_VALUE                    (TE_TIMER_OVERFLOW_VAL - 40 * TE)

// Values that the timer will be compared to to check the elapsed time. During a
// DALI frame reception at each bus transition the software arms the Te timer
// to trigger an interrupt after 4 Te. If this happens, it will be considered a
// timeout (and unless this is the last bit reception will be aborted and an
// error will be signalled) since all delays within a frame should be at most 2 Te.
// If the external interrupt is triggered during this 4 Te period, the timer
// register is compared to these 4 values and if it falls within [TE_MIN, TE_MAX]
// a single Te is considered to have elapsed, whereas if it falls within
// [TE2_MIN, TE2_MAX], 2 Te is considered to have elapsed. Otherwise, a reception
// error is signalled.
#define TE_MIN                          (TE4_VALUE + 0.5 * TE)
#define TE_MAX                          (TE4_VALUE + 1.3 * TE)
#define TE2_MIN                         (TE4_VALUE + 0.7 * 2 * TE)
#define TE2_MAX                         (TE4_VALUE + 1.3 * 2 * TE)

// During transmission on the DALI bus the firmware does collision detection. In
// doing so, it expects to see on the RX pin the value that is set on the TX pin.
// However, the bus takes a while to respond (depending on the driver strength,
// bus capacitance and power supply). Also during the time that we're driving the
// line we don't expect to see any transitions.
// The implementation is that we define a 150us window during which we expect a
// transition if one was generated. If no such transition occurs, or transitions
// occur outside this window, a collision is signalled.
#define TE_TRANSITION_VALID_MAX         (TE_VALUE + 450)

// Maximum number of components an "arbitrary" digital waveform may have. The
// same vector is used both for transmission and reception of such signals.
#define TIMING_LIST_SIZE                120

// DALI bus decoded data is put in a circular buffer such that the application
// can pick it up at its own pace. This is the size of this buffer.
#define RX_QUEUE_SIZE                   20



// DALI protocol state machine flags. Most of these get reset at the start of each
// frame and are updated as the machine advances. At the end of the frame (or
// frames, since the machine doesn't return to ST_IDLE between a forward frame
// and its associated backward frame) it digests these flags into fewer bits
// which it records along with the received/transmitted data.
// Most of these flags are used to communicate information between states.
typedef union
{
    uint16_t All;
    struct
    {
        unsigned txError                        : 1;    // Transmission error
        unsigned txDone                         : 1;    // Transmission done
        unsigned rxError                        : 1;    // Reception error
        unsigned rxDone                         : 1;    // Reception done
        unsigned rxTimingListDone               : 1;    // Finished receiving a waveform capture from the DALI bus
        unsigned configRxTimingList             : 1;    // Selects waveform capture instead of DALI protocol decoding
        unsigned rxTimingError                  : 1;    // Timing error on reception
        unsigned rxBackward                     : 1;    // Backward frame received
        unsigned rxForward                      : 1;    // Forward frame received
        unsigned cableConnected                 : 1;    // DALI cable connected to the board
        unsigned rxFromIdle                     : 1;    // Started receiving frame from the ST_IDLE state
        unsigned rxFromWaitForBackframe         : 1;    // Started receiving frame from ST_WAIT_FOR_BACKFRAME
        unsigned txThisDevice                   : 1;    // This is the device that put the forward frame on the bus
        unsigned rxPacketTimeValid              : 1;    // Received packet with correct timing
        unsigned                                : 2;
    };
}flags_type;



// The state of the DALI state machine. This can be one of ST_IDLE, ST_SEND_DATA,
// etc. These are defined in dali.h since the application software needs to be
// able to check these states as well, mainly for debugging purposes.
uint8_t daliState;

// Variables used by the state machine
volatile uint16_t forwardFrame;
volatile uint16_t txPacket;
volatile uint16_t rxPacket;
volatile uint16_t rxPacketTime;
volatile uint8_t halfBitNumber;
volatile uint8_t transitionValid;
volatile uint8_t rxBackwardFrame;
volatile uint8_t rxPacketLen;

// DALI protocol state machine flags variable
flags_type DALIFlags;

// Vector containing the components of either the captured waveform or the one
// to be sent
uint16_t timingList[TIMING_LIST_SIZE] = {0xFFFF};

// Cursor withing this vector
volatile uint8_t dataTimingListCounter;

// Data circular buffer where the DALI stack fills in received data. The
// application can then grab the data from here
uint16_t forwardData[RX_QUEUE_SIZE];
uint16_t backwardFrameDelay[RX_QUEUE_SIZE];
uint8_t backwardData[RX_QUEUE_SIZE];
bundle_flags_type flagsData[RX_QUEUE_SIZE];
volatile uint8_t dataR, dataW;


/***********************Local function definitions*****************************/

// Clear most flags such that the new frame gets a fresh start. The cable
// connected and configuration flags remain unchanged
void DALIClearFlags(void);

// Once the machine is ready receiving data, it will use this function to append
// the data to the queue. The application can check if this queue is empty and
// if there is data ready can then retrieve it.
void DALIAppendToQueue(void);

/***********************Local function definitions*(END)***********************/


/*************************Function implementations*****************************/


inline void DALIInit(void)
{
    // Initialize the bus to idle state
    DALI_TX = DALI_HI;

    // Set tristate bits. Rx is input, Tx is output
    DALI_TX_TRIS = 0;
    DALI_RX_TRIS = 1;

    // 0b00100000
    // 0b00xxxxxx - FOSC/4
    // 0bxx10xxxx - Prescale 1:4
    // 0bxxxx0xxx - Dedicated osc disabled
    // 0bxxxxx0xx - Sync
    // 0bxxxxxx0x - Unimplemented
    // 0bxxxxxxx0 - Timer stopped
    T1CON = 0b00100000;

    // Clear interrupt flag
    TE_TIMER_IF = 0;

    // Enable timer1 interrupt and keep it enabled for the whole duration of the
    // application. Will turn timer on/off as needed
    TMR1IE = 1;

    // Clear external interrupt flag
    INT_IF = 0;

    // Enable external interrupt and keep it enabled for the whole duration of
    // the application. We're always on the look-out for things happening on the
    // DALI bus.
    INTE = 1;

    // Initialize state machine
    daliState = ST_IDLE;

    // And its flags
    DALIFlags.All = 0;

    // Start with the circular data buffer empty.
    dataR = 0;
    dataW = 0;

    // Based on the state of the bus at power-up we set the external interrupt
    // such that it will trigger on the first bus transition.s
    if (DALI_RX == DALI_LO)
    {
        INT_DALI_RISING();
    }
    else
    {
        INT_DALI_FALLING();
    }
}


inline void DALIInterrupt(void)
{
    static uint16_t intValue;
    static uint8_t cableDisconnectCounter = 0;

    // Handle bus connection/disconnection. If the  bus is disconnected for 20ms
    // or more, the application can know.
    if (TICK_TIMER_IF != 0)
    {
        TICK_TIMER_IF = 0;

        // Check if cable is connected
        if (DALI_RX == DALI_LO)
        {
            if (cableDisconnectCounter > 0)
            {
                cableDisconnectCounter -= 1;
            }
        }
        else
        {
            cableDisconnectCounter = 20;
        }

        if (cableDisconnectCounter == 0)
        {
            DALIFlags.cableConnected = 0;
        }
        else
        {
            DALIFlags.cableConnected = 1;
        }

    }

    // The timer and external interrupts run the machine both for reception and
    // transmission of data.

    if (TE_TIMER_IF != 0)
    {
        // Te timer interrupt
        TE_TIMER_IF = 0;
        if (daliState == ST_SEND_DATA_TIMING_LIST)
        {
            // We just need to keep loading values from the buffer into the timer
            // register and wait for the next interrupt. When we hit 0xFFFF we
            // need to stop. Since the values in the buffer represent the times
            // until the bus needs to change state, once we load a new value we
            // aslo toggle the bus line.
            dataTimingListCounter++;
            if (timingList[dataTimingListCounter] != 0xFFFF)
            {
                TE_TIMER_VAL += timingList[dataTimingListCounter];
                DALI_TX ^= 1;
            }
            else
            {
                // We finished sending the arbitrary waveform. We now need to
                // listen on the bus for any reply, since the user might have
                // constructed a waveform that corresponds to a valid DALI
                // command. Therefore we need to make sure the bus is left in an
                // idle state.
                // In order to allow the machine to also check the timing of the
                // backward frame and to not duplicate code, after we've
                // finished sending the waveform we switch to the ST_SEND_DATA
                // state and tell the machine it still needs to transmit the
                // stop bits (halfBitNumber = 35). From then on, the machine
                // won't care if what has been sent was actual DALI data or an
                // arbitrary waveform specified by the user.
                // The transitionValid variable needs to be set appropriately for
                // the ST_SEND_DATA state such that it won't trigger an error.
                if (DALI_TX == DALI_HI)
                {
                    transitionValid = 1;
                }
                else
                {
                    DALI_TX = DALI_HI;
                    transitionValid = 0;
                }
                TE_TIMER_VAL += TE_VALUE;
                halfBitNumber = 35;
                daliState = ST_SEND_DATA;
            }
        }
        else if (daliState == ST_SEND_DATA)
        {
            // Send data. This state uses the halfBitNumber to track its
            // progress and transitionValid to check if since the last half bit
            // the bus behaved correctly (had just the expected transitions at
            // the expected times)
            if (transitionValid != 1)
            {
                // Erroneous transition. Either none or too many transitions
                // happend on the RX line. Signal error (and type) and reset
                // state machine. Move to ST_BUS_IGNORE which ignores the bus
                // for TE_ERR_VALUE. This way if something was causing trouble
                // on the bus we don't want to be able to add to it, nor to keep
                // informing the upper layers that data is available.
                DALI_TX = DALI_HI;
                DALIFlags.txError = 1;
                DALIFlags.txDone = 1;
                TE_TIMER_VAL += TE_ERR_VALUE;
                daliState = ST_BUS_IGNORE;
                DALIAppendToQueue();
                return;
            }
            // Transition ok, go forward
            switch (halfBitNumber)
            {
                case 1:
                    // START BIT, 2nd half
                    transitionValid = 0;
                    // The timer keeps running, at this current time it couldn't
                    // be too far away from 0, but in order to increase the accuracy
                    // of the baud rate, we add to this value instead of replacing it.
                    // This has the advantage of removing most of the delay from the
                    // ISR vectoring up to this point
                    TE_TIMER_VAL += TE_VALUE;
                    DALI_TX = DALI_HI;
                    break;
                case 2:
                    // DATA BIT, 1st half
                    // Now is the time to set the DALI bus to the proper level for
                    // bit 0 of our data. We also need to set transitionValid according
                    // to the transitioning from the start bit, bus is currently
                    // at DALI_HI:
                    // DALI_TX = DALI_LO ==> transitionValid = 0
                    // DALI_TX = DALI_HI ==> transitionValid = 1 which will make
                    // our test pass the check and will enable the state machine
                    // to count any (unexpected) transitions as well
                    TE_TIMER_VAL += TE_VALUE;
                    if ((txPacket & 0x8000) != 0)
                    {
                        // Send a '1' bit, which starts with a DALI_LO
                        DALI_TX = DALI_LO;
                        transitionValid = 0;
                    }
                    else
                    {
                        // Send a '0' bit, which starts with a DALI_HI
                        DALI_TX = DALI_HI;
                        transitionValid = 1;
                    }
                    break;
                case 34:
                    // STOP BIT 1, 1st half
                    TE_TIMER_VAL += TE_VALUE;
                    DALI_TX = DALI_HI;
                    if ((txPacket & 0x8000) != 0)
                    {
                        // Having last transmitted a '1', the bus was already
                        // in state DALI_HI, so we won't expect a new transition.
                        transitionValid = 1;
                    }
                    else
                    {
                        // We need to transition to DALI_HI
                        transitionValid = 0;
                    }
                    break;
                case 35:
                case 36:
                case 37:
                    // STOP BIT 1, 2nd half
                    // STOP BIT 2, 1st half
                    // STOP BIT 2, 2nd half
                    TE_TIMER_VAL += TE_VALUE;
                    DALI_TX = DALI_HI;
                    // No transitions should occur, we should stay put
                    transitionValid = 1;
                    break;
                case 38:
                    // Transmission finalized OK. Start waiting
                    // for a potential backframe. Keep timer running
                    TE_TIMER_VAL += TE22_VALUE;
                    rxPacketTime = 0;
                    daliState = ST_WAIT_FOR_BACKFRAME;
                    break;
                default:
                    // This case handles all the in-between (data) bits, with a
                    // distinction between 1st half (even halfBitNumber) and 2nd
                    // half (odd halfBitNumber) of a bit.
                    TE_TIMER_VAL += TE_VALUE;
                    if ((halfBitNumber & 0x01) == 0)
                    {
                        // halfBitNumber is even, we are in the 1st half of the bit.
                        switch (txPacket & 0xC000)
                        {
                            case 0x0000:
                                // Last bit sent 0, we need to send 0
                                DALI_TX = DALI_HI;
                                transitionValid = 0;
                                break;
                            case 0x4000:
                                // Last bit sent 0, we need to send 1
                                DALI_TX = DALI_LO;
                                transitionValid = 1;
                                break;
                            case 0x8000:
                                // Last bit sent 1, we need to send 0
                                DALI_TX = DALI_HI;
                                transitionValid = 1;
                                break;
                            case 0xC000:
                                // Last bit sent 1, we need to send 1
                                DALI_TX = DALI_LO;
                                transitionValid = 0;
                                break;
                        }
                        txPacket <<= 1;
                    }
                    else
                    {
                        // halfBitNumber is odd, 2nd half of the bit.
                        if ((txPacket & 0x8000) != 0)
                        {
                            // We're currently sending a '1'
                            DALI_TX = DALI_HI;
                        }
                        else
                        {
                            // We're currently sending a '0'
                            DALI_TX = DALI_LO;
                        }
                        // Always expect transition in the middle of the bit
                        transitionValid = 0;
                    }
                    break;
            }
            // Move on in state ST_DATA, by incrementing the halfbit we're transmitting
            halfBitNumber++;
        }
        else if (daliState == ST_WAIT_FOR_BACKFRAME)
        {
            // Time-out occured while waiting for backframe
            if (DALIFlags.txThisDevice == 1)
            {
                DALIFlags.txDone = 1;
            }
            else
            {
                DALIFlags.rxDone = 1;
            }
            TE_TIMER_VAL += TE4_VALUE;
            daliState = ST_PRE_IDLE;
            DALIAppendToQueue();
        }
        else if (daliState == ST_WAIT_AFTER_BACKFRAME)
        {
            // Time-out occured while waiting after backframe. This is ok, nobody
            // should have bothered the bus in this period.
            TE_TIMER_VAL += TE4_VALUE;
            daliState = ST_PRE_IDLE;
            DALIAppendToQueue();
        }
        else if (daliState == ST_BUS_IGNORE)
        {
            // Safeguard period over, we can go to ST_PRE_IDLE.
            TE_TIMER_VAL += TE4_VALUE;
            daliState = ST_PRE_IDLE;
        }
        else if (daliState == ST_PRE_IDLE)
        {
            // Safeguard period over, we can go back to ST_IDLE.
            TE_TIMER_ON = 0;
            daliState = ST_IDLE;
        }
        else if (daliState == ST_RECEIVE_DATA)
        {
            // Time-out. Check first if we're at a final bit (got 8 or 16 bits)
            // and the last bit is a 1 or a 0. We need to add an extra delay
            // of 1 Te if this is the final bit of a packet and the last bit was
            // a 1, just to make sure that we exit ST_RECEIVE_DATA after the last
            // stop bit has been fully received.
            // Otherwise, if this is not the last bit of a frame, we need to
            // signal an error.
            if (((rxPacketLen == 8) || (rxPacketLen == 16)) && ((rxPacket & 0x01) != 0x00))
            {
                // The last bit we've received is very likely the last one in the
                // transmission, and it is a '1'. Thus, we need to add a 1 Te
                // delay.
                // Keep TE_TIMER_ON unmodified, reload timer with 1 Te.
                daliState = ST_RECEIVE_DATA_EXTRA_TE;
                TE_TIMER_VAL = TE_VALUE;
            }
            else
            {
                // We've finished receiving some data.
                if (rxPacketLen == 8)
                {
                    // What we've just got is a backward frame, as it only has
                    // 8 bits. If we started receiving it out of the blue (from
                    // ST_IDLE) this means the backframe came too late after its
                    // forward frame, so we need to set the rxTimingError bit.
                    DALIFlags.rxBackward = 1;
                    rxBackwardFrame = rxPacket;

                    if (DALIFlags.txThisDevice == 1)
                    {
                        DALIFlags.txDone = 1;
                    }

                    if (DALIFlags.rxFromIdle == 1)
                    {
                        DALIFlags.rxTimingError = 1;
                    }

                    // The halfBitNumber internal state gives the position within
                    // the last received bit. This is important since we need to
                    // know the current state of the DALI_RX line. These two values
                    // are the ones where the line returns to DALI_HI.
                    if ((halfBitNumber != 2) && (halfBitNumber != 4))
                    {
                        DALIFlags.rxError = 1;
                    }

                    TE_TIMER_VAL += TE22_VALUE;
                    DALIFlags.rxDone = 1;
                    daliState = ST_WAIT_AFTER_BACKFRAME;
                }
                else if (rxPacketLen == 16)
                {
                    // What we've just got is a forward frame, as it has 16 bits.
                    // If we started receiving it while waiting for a backframe
                    // (from ST_WAIT_FOR_BACKFRAME) this means that a master did
                    // not respect the minimum time between frames. Thus we need
                    // to set the rxTimingError bit.
                    DALIFlags.rxForward = 1;
                    forwardFrame = rxPacket;

                    if (DALIFlags.rxFromWaitForBackframe == 1)
                    {
                        DALIFlags.rxTimingError = 1;
                    }

                    if ((halfBitNumber != 2) && (halfBitNumber != 4))
                    {
                        DALIFlags.rxError = 1;
                    }

                    DALIFlags.rxFromWaitForBackframe = 0;
                    DALIFlags.rxFromIdle = 0;
                    TE_TIMER_VAL += TE22_VALUE;
                    rxPacketTime = 0;
                    daliState = ST_WAIT_FOR_BACKFRAME;
                }
                else
                {
                    // We've received a number of bits that don't make up a whole
                    // frame and we need to signal an error.
                    DALIFlags.rxError = 1;
                    DALIFlags.rxDone = 1;
                    TE_TIMER_VAL += TE_ERR_VALUE;
                    daliState = ST_BUS_IGNORE;
                    DALIAppendToQueue();
                }
            }
        }
        else if (daliState == ST_RECEIVE_DATA_EXTRA_TE)
        {
            // Time-out. If we didn't have any transition during this last wait
            // (which is safe to assume as such a transition would have stopped
            // the transmission and signaled an error already) we can check if
            // the received frame is valid.
            if (rxPacketLen == 8)
            {
                // Backward frame, same checks as during ST_RECEIVE_DATA
                DALIFlags.rxBackward = 1;
                rxBackwardFrame = rxPacket;
                if (DALIFlags.txThisDevice == 1)
                {
                    DALIFlags.txDone = 1;
                }

                if (DALIFlags.rxFromIdle == 1)
                {
                    DALIFlags.rxTimingError = 1;
                }

                if ((halfBitNumber != 2) && (halfBitNumber != 4))
                {
                    DALIFlags.rxError = 1;
                }


                TE_TIMER_VAL += TE22_VALUE;
                DALIFlags.rxDone = 1;
                daliState = ST_WAIT_AFTER_BACKFRAME;
            }
            else if (rxPacketLen == 16)
            {
                // Forward frame, same checks as during ST_RECEIVE_DATA
                DALIFlags.rxForward = 1;
                forwardFrame = rxPacket;
                if (DALIFlags.rxFromWaitForBackframe == 1)
                {
                    DALIFlags.rxTimingError = 1;
                }

                if ((halfBitNumber != 2) && (halfBitNumber != 4))
                {
                    DALIFlags.rxError = 1;
                }


                DALIFlags.rxFromWaitForBackframe = 0;
                DALIFlags.rxFromIdle = 0;
                TE_TIMER_VAL += TE22_VALUE;
                rxPacketTime = 0;
                daliState = ST_WAIT_FOR_BACKFRAME;
            }
            else
            {
                // Error condition, same checks as during ST_RECEIVE_DATA
                DALIFlags.rxError = 1;
                DALIFlags.rxDone = 1;
                TE_TIMER_VAL += TE_ERR_VALUE;
                daliState = ST_BUS_IGNORE;
                DALIAppendToQueue();
            }
        }
        else if (daliState == ST_RECEIVE_DATA_TIMING_LIST)
        {
            // Time-out. We'll stop here, place the end-of-frame marker (0xFFFF)
            // and send the data to the PC.
            timingList[dataTimingListCounter] = 0xFFFF;
            TE_TIMER_ON = 0;
            daliState = ST_IDLE;
            dataTimingListCounter = 0;
            DALIFlags.rxTimingListDone = 1;
        }
    }

    if (INT_IF != 0)
    {
        // DALI BUS changed. This could either be an expected transition.. or not
        INT_IF = 0;

        // Every falling edge is followed by a rising edge and the other way around
        INT_DALI_TOGGLE();

        if ((daliState == ST_IDLE) || (daliState == ST_PRE_IDLE))
        {
            // Something unexpected happened on the bus.
            // Time-out should be dealt with according to the next state. If we
            // will go to ST_RECEIVE_DATA_TIMING_LIST then the time-out will
            // mean the end of the frame. If, however, we were expecting data
            // with ST_RECEIVE_DATA, this will signal a bus error.
            // if DALIFlags.configRxTimingList == 0 go to ST_RECEIVE_DATA, else
            // go to ST_RECEIVE_DATA_TIMING_LIST.

            // We should make sure we're not driving our own line low
            DALI_TX = DALI_HI;
            if (INT_DALI_IS_FALLING())
            {
                // After toggling we now have a falling edge detect enabled, which
                // means that we got here by a rising edge event. This should normally
                // only happen after the cable has just been connected. Thus we
                // stay put..
            }
            else
            {
                // Either some data comes along on the bus or the cable has been
                // disconnected. We can't know, so we're assuming the former.
                // Based on our configRxTimingList we need to start one of the
                // two reception mechanisms.
                if (DALIFlags.configRxTimingList == 0)
                {
                    // Prepare for data reception and decoding. The time-out in
                    // this case is 4 Te, after which we'll signal an error if no
                    // other transitions occur.
                    rxPacket = 0;
                    rxPacketLen = 0;
                    halfBitNumber = 0;
                    DALIFlags.rxDone = 0;
                    DALIFlags.rxFromIdle = 1;
                    TE_TIMER_VAL = TE4_VALUE;
                    daliState = ST_RECEIVE_DATA;
                }
                else
                {
                    // Prepare for waveform capture. The time-out here is the
                    // timer period (~21.8ms), since the user may want to capture
                    // more than one DALI frame.
                    TE_TIMER_VAL = 0; // time-out = timer period (~21.8ms)
                    daliState = ST_RECEIVE_DATA_TIMING_LIST;
                    dataTimingListCounter = 0;
                    DALIFlags.rxTimingListDone = 0;
                }
                // If we got here from ST_PRE_IDLE timer may have expired
                TE_TIMER_IF = 0;
                TE_TIMER_ON = 1;
            }
        }
        else if (daliState == ST_WAIT_AFTER_BACKFRAME)
        {
            // Any transition during ST_WAIT_AFTER_BACKFRAME will generate a
            // timing error and will move us to the data reception state.
            rxPacket = 0;
            rxPacketLen = 0;
            halfBitNumber = 0;
            DALIFlags.rxDone = 0;
            DALIFlags.rxTimingError = 1;
            TE_TIMER_VAL = TE4_VALUE;
            daliState = ST_RECEIVE_DATA;
            TE_TIMER_ON = 1;
        }
        else if (daliState == ST_SEND_DATA)
        {
            // Transitions during data sending may or may not be allowed. If
            // allowed, these should happen within a timing window of 150us after
            // we have set a value on the data bus.
            // States set transitionValid to 0 when expecting a transition in
            // the following 150us and to 1 when not expecting any transition
            // until the timer expires. When the timer expires this variable is
            // tested and if it differs from 1, an error is signalled.
            // The mechanism here handles the two situations. If transitionValid
            // is 0 we check if we are within the allowed window and increment
            // transitionValid if this is true.
            // If transitionValid is 1 (i.e. we're not expecting any transitions)
            // we increment it unconditionally. This means that any value
            // other than 1 will mean an error has occured.
            // For example, if the expected transition does not come within the
            // window but comes at a later time, transitionValid will remain
            // fixed at the value 0.
            if (transitionValid == 0)
            {
                intValue = TE_TIMER_VAL;
                if (intValue > TE_VALUE && intValue < TE_TRANSITION_VALID_MAX)
                {
                    transitionValid++;
                }
            }
            else
            {
                transitionValid++;
            }
        }
        else if (daliState == ST_WAIT_FOR_BACKFRAME)
        {
            // Transition during this state signals the start of a backframe,
            // but it could be that this is a new forward frame being sent by a
            // Control Device that doesn't respect the inter-frame timing.
            // In any case the start of this frame is timed against the beginning
            // of the state to make sure the backframe comes within its allowed
            // time slot (or signal an error otherwise).
            // The DALIFlags.rxFromWaitForBackframe bit will tell later states
            // that the current frame started being received at this point. If
            // it turns out this is a backframe, this is a requirement, whereas
            // if it turns out to be a forward frame, this will flag an error.
            rxPacketTime = TE_TIMER_VAL;
            DALIFlags.rxPacketTimeValid = 1;
            if (rxPacketTime <= (TE22_VALUE + 7 * TE))
            {
                DALIFlags.rxTimingError = 1;
            }
            rxPacket = 0;
            rxPacketLen = 0;
            halfBitNumber = 0;
            DALIFlags.rxFromWaitForBackframe = 1;
            TE_TIMER_VAL = TE4_VALUE;
            daliState = ST_RECEIVE_DATA;
            // Timer may have overflowed while we were servicing this
            // interrupt
            TE_TIMER_IF = 0;
        }
        else if (daliState == ST_RECEIVE_DATA)
        {
            // Bus transition during data reception. We need to reload the timer
            // with a new value (such that it won't time-out and that our
            // measurements will make sense)
            // The current value of the timer needed to distinguish between the
            // data being received.
            intValue = TE_TIMER_VAL;
            TE_TIMER_VAL = TE4_VALUE;
            switch (halfBitNumber)
            {
                case 0:
                    // halfBitNumber == 0 means "this should be the middle of the
                    // start bit". Thus a Te now means we've got it ok, while anything
                    // else means an error has occured.
                    if ((intValue >= TE_MIN) && (intValue <= TE_MAX))
                    {
                        halfBitNumber = 2;
                    }
                    else
                    {
                        // Signal reception error if the bit timing isn't
                        // respected.
                        DALIFlags.rxError = 1;
                    }
                    break;
                case 1:
                    // halfBitNumber == 1 means "we were in the first part of receiving
                    // a 1". Thus a Te now means we've got the 1 ok, while anything else
                    // means an error has occured.
                    if ((intValue >= TE_MIN) && (intValue <= TE_MAX))
                    {
                        // Middle of a '1' bit
                        rxPacket <<= 1;
                        rxPacket |= 1;
                        rxPacketLen++;
                        halfBitNumber = 2;
                    }
                    else
                    {
                        // Signal reception error if the bit timing isn't
                        // respected.
                        DALIFlags.rxError = 1;
                    }
                    break;
                case 2:
                    // halfBitNumber == 2 means "we were in the middle of receiving
                    // a 1". Thus a Te now means we're getting another 1 and this is
                    // the beginning, a 2 Te means we're getting a 0 and this is
                    // the middle.
                    if ((intValue >= TE_MIN) && (intValue <= TE_MAX))
                    {
                        // 1st half of a '1' bit
                        halfBitNumber = 1;
                    }
                    else if ((intValue >= TE2_MIN) && (intValue <= TE2_MAX))
                    {
                        // Middle of a '0' bit
                        rxPacket <<= 1;
                        rxPacketLen++;
                        halfBitNumber = 3;
                    }
                    else
                    {
                        // Signal reception error if the bit timing isn't
                        // valid.
                        DALIFlags.rxError = 1;
                    }
                    break;
                case 3:
                    // halfBitNumber == 3 means "we were in the middle of receiving
                    // a 0". Thus a Te now means we're getting another 0 and this is
                    // the beginning, a 2 Te means we're getting a 1 and this is
                    // the middle.
                    if ((intValue >= TE_MIN) && (intValue <= TE_MAX))
                    {
                        // 1st half of a '0' bit
                        halfBitNumber = 4;
                    }
                    else if ((intValue >= TE2_MIN) && (intValue <= TE2_MAX))
                    {
                        // Middle of a '1' bit
                        rxPacket <<= 1;
                        rxPacket |= 1;
                        rxPacketLen++;
                        halfBitNumber = 2;
                    }
                    else
                    {
                        // Signal reception error if the bit timing isn't
                        // respected.
                        DALIFlags.rxError = 1;
                    }
                    break;
                case 4:
                    // halfBitNumber == 4 means "we were in the first part of receiving
                    // a 0". Thus a Te now means we've got the 0 ok, while 2 Te means
                    // an error has occured.
                    if ((intValue >= TE_MIN) && (intValue <= TE_MAX))
                    {
                        // 1st half of a '0' bit received OK
                        rxPacket <<= 1;
                        rxPacketLen++;
                        halfBitNumber = 3;
                    }
                    else
                    {
                        // Signal reception error if the bit timing isn't
                        // respected.
                        DALIFlags.rxError = 1;
                    }
                    break;
            }
        }
        else if (daliState == ST_RECEIVE_DATA_EXTRA_TE)
        {
            // Transition during the last half of the 2nd stop bit, which is not
            // supposed to happen. Signal RX error.
            DALIFlags.rxError = 1;
        }
        else if (daliState == ST_RECEIVE_DATA_TIMING_LIST)
        {
            // A transition during the capture of the waveform means we need to
            // save the current timer value in the buffer (if there's space for
            // it) and start the next measurement.
            intValue = TE_TIMER_VAL;
            TE_TIMER_VAL = 0;
            if (dataTimingListCounter < TIMING_LIST_SIZE - 1)
            {
                timingList[dataTimingListCounter] = intValue;
                dataTimingListCounter++;
            }
        }
    }
}

/*********Arbitrary waveform setup and transmission*********/

void DALIResetTimingList(void)
{
    dataTimingListCounter = 0;
}

void DALILoadTimingListValue(uint16_t val)
{
    if (dataTimingListCounter < TIMING_LIST_SIZE - 1)
    {
        timingList[dataTimingListCounter] = val;
        if (val != 0xFFFF)
        {
            dataTimingListCounter++;
        }
    }
    else
    {
        timingList[dataTimingListCounter] = 0xFFFF;
    }
}

void DALISendDataTimingList(void)
{
    dataTimingListCounter = 0;
    daliState = ST_SEND_DATA_TIMING_LIST;
    TE_TIMER_VAL = timingList[dataTimingListCounter];
    DALI_TX = DALI_LO;
    TE_TIMER_ON = 1;
}


/*********Arbitrary waveform setup and transmission**(END)**/



/*********Arbitrary waveform reception*********/

void DALISetReceiveTimingList(uint8_t val)
{
    if (val == 0)
    {
        DALIFlags.configRxTimingList = 0;
    }
    else
    {
        DALIFlags.configRxTimingList = 1;
    }
}

uint8_t DALIReceiveTimingListDone(void)
{
    if (DALIFlags.rxTimingListDone == 0)
    {
        return 0xFF;
    }
    else
    {
        return dataTimingListCounter;
    }
}

uint16_t DALIReadTimingListValue(void)
{
    uint16_t retVal;
    retVal = timingList[dataTimingListCounter];
    if (retVal != 0xFFFF)
    {
        dataTimingListCounter++;
    }
    else
    {
        DALIFlags.rxTimingListDone = 0;
    }
    return retVal;
}

/*********Arbitrary waveform reception**(END)**/




/*********DALI Data transmission and reception*********/

uint8_t DALISendData(uint8_t addr, uint8_t val)
{
    if (daliState == ST_IDLE)
    {
        daliState = ST_SEND_DATA;
        DALIFlags.txThisDevice = 1;
        transitionValid = 0;
        halfBitNumber = 1;
        txPacket = addr << 8 | val;
        forwardFrame = txPacket;
        // expect this transition to result in an input interrupt
        INT_DALI_FALLING();
        TE_TIMER_VAL = TE_VALUE;
        DALI_TX = DALI_LO;
        TE_TIMER_ON = 1;
        return 0;
    }
    else
    {
        return 1;
    }
}

uint8_t DALIDataAvailable(void)
{
    return dataR != dataW;
}

uint8_t DALIReceiveDataFlags(void)
{
    dataR = (dataR + 1) % RX_QUEUE_SIZE;
    return flagsData[dataR].flagsByte;
}

uint16_t DALIReceiveDataForward(void)
{
    return forwardData[dataR];
}

uint8_t DALIReceiveDataBackward(void)
{
    return backwardData[dataR];
}

uint16_t DALIReceiveDataTiming(void)
{
    return backwardFrameDelay[dataR];
}

/*********DALI Data transmission and reception**(END)**/



/*********DEBUG Functions*********/

inline uint16_t DALIReadFlags(void)
{
    return DALIFlags.All;
}

inline uint8_t DALIReadState(void)
{
    return daliState;
}

/*********DEBUG Functions**(END)**/



void DALIClearFlags(void)
{
    uint8_t a, b;
    a = DALIFlags.cableConnected;
    b = DALIFlags.configRxTimingList;
    DALIFlags.All = 0;
    DALIFlags.cableConnected = a;
    DALIFlags.configRxTimingList = b;
}



void DALIAppendToQueue(void)
{
    // When the state machine finishes receiving data or encounters an error it
    // uses this function to submit the data to the buffer such that the
    // application can grab it. The flags that are presented to the application
    // are a digested version of the state machine flags at the moment this
    // function is called.
    
    // Make sure we don't start overwriting old data that hasn't yet been read by
    // the application.
    if ((dataW + 1) % RX_QUEUE_SIZE != dataR)
    {
        // Advance writing head on the circular buffer
        dataW = (dataW + 1) % RX_QUEUE_SIZE;

        // Fill data in
        forwardData[dataW] = forwardFrame;
        backwardFrameDelay[dataW] = rxPacketTime;
        backwardData[dataW] = rxBackwardFrame;

        // Fill flags in
        flagsData[dataW].txThisDevice = DALIFlags.txThisDevice;
        flagsData[dataW].txError = DALIFlags.txError;
        flagsData[dataW].rxError = DALIFlags.rxError;
        flagsData[dataW].rxTimingError = DALIFlags.rxTimingError;
        flagsData[dataW].backwardFrameDelayValid = DALIFlags.rxPacketTimeValid;
        flagsData[dataW].backwardFrameValid = DALIFlags.rxBackward;
        flagsData[dataW].forwardFrameValid = DALIFlags.rxForward | (DALIFlags.txDone && !DALIFlags.txError);

        // Clear machine flags such that the next frame will not inherit garbage.
        DALIClearFlags();
    }
}


/*************************Function implementations*(END)***********************/
