/********************************************************************
 *
 * Copyright (C) 2014 Microchip Technology Inc. and its subsidiaries
 * (Microchip).  All rights reserved.
 *
 * You are permitted to use the software and its derivatives with Microchip
 * products. See the license agreement accompanying this software, if any, for
 * more info about your rights and obligations.
 *
 * SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR
 * PURPOSE. IN NO EVENT SHALL MICROCHIP, SMSC, OR ITS LICENSORS BE LIABLE OR
 * OBLIGATED UNDER CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH
 * OF WARRANTY, OR OTHER LEGAL EQUITABLE THEORY FOR ANY DIRECT OR INDIRECT
 * DAMAGES OR EXPENSES INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES, OR OTHER SIMILAR COSTS. To the fullest
 * extend allowed by law, Microchip and its licensors liability will not exceed
 * the amount of fees, if any, that you paid directly to Microchip to use this
 * software.
 *************************************************************************
 *
 *                           dali.h
 *
 * About:
 *  Header file for the DALI protocol implementation
 *
 *
 * Hardware:
 *  USB to DALI Interface - DM160215
 *
 *
 * Author            Date            Ver     Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Mihai Cuciuc      18 October 2013  1.00    Implemented DALI protocol
 *
 *************************************************************************/

#ifndef DALI_H
#define	DALI_H

#include <xc.h>
#include <stdint.h>

enum
{
    ST_IDLE,                        // DALI state machine idle
    ST_SEND_DATA_TIMING_LIST,       // Sending a digital waveform dictated by the PC application
    ST_SEND_DATA,                   // Sending a DALI command
    ST_RECEIVE_DATA_TIMING_LIST,    // Capturing a digital waveform from the DALI bus
    ST_RECEIVE_DATA,                // Receiving data from the DALI bus
    ST_RECEIVE_DATA_EXTRA_TE,       // Based on the last RX'd bit, we may need to wait an extra Te
    ST_WAIT_FOR_BACKFRAME,          // Waiting for Control Gears to respond, if they have anything to say
    ST_WAIT_AFTER_BACKFRAME,        // There's a minimum delay after having received data until we can send again
    ST_PRE_IDLE,                    // Safeguard against transmitting a second packet too early after the last one
    ST_BUS_IGNORE                   // After detecting a bus error, we will ignore the bus for a predetermined time to make sure the error went away
};

// The data buffer contains the forward frame, the backward frame and the delay
// between these two. This set of flags give the validity of each of these data
// as well as other information such as wether this is the device that sent the
// forward frame or it's a packet sniffed from the bus.
typedef union
{
    uint8_t flagsByte;
    struct
    {
        unsigned forwardFrameValid                   : 1;   // Forward frame valid
        unsigned backwardFrameValid                  : 1;   // Backward frame valid
        unsigned backwardFrameDelayValid             : 1;   // Delay between forward and backward frames valid
        unsigned txThisDevice                        : 1;   // This device transmitted the forward frame
        unsigned txError                             : 1;   // Error on transmission
        unsigned rxTimingError                       : 1;   // Reception of data at illegal time
        unsigned rxError                             : 1;   // Error on reception
        unsigned                                     : 1;
    };
}bundle_flags_type;


/**********************Public function definitions*****************************/

// Initialize DALI stack.
inline void DALIInit(void);

// This needs to be called from the ISR unconditionally. It checks and clears
// the Tick timer, Te timer and External interrupt flags.
inline void DALIInterrupt(void);




/*********Arbitrary waveform setup and transmission*********/

// Move data pointer to the beginning of the timing list, to start writing to
// it from the beginning.
void DALIResetTimingList(void);

// Append to the timing list value val. This timing list will make up the digital
// waveform to be sent on the DALI bus. The value val is what will be loaded into
// the timer register and as such, the delay will be 65536 - val timer clocks.
// If there's no room left for the data, an end-of-list marker is appended and
// the provided value is discarded.
void DALILoadTimingListValue(uint16_t val);

// Once the timing list is prepared, a call to this function will start putting
// this waveform on the DALI bus. The whole process looks something like this:
//      DALIResetTimingList();
//      DALILoadTimingListValue(64286);
//      DALILoadTimingListValue(64286);
//      DALILoadTimingListValue(64286);
//      DALISendDataTimingList();
// This will send a signal on the DALI bus that looks like this
// .........______    ____    ________........
//                |___|   |___|
// with the width of the pulses equal to 1250 timer clocks. The timer is configured
// for 3MHz clock, so each of these pulses will be roughly 416us long.
void DALISendDataTimingList(void);

/*********Arbitrary waveform setup and transmission**(END)**/



/*********Arbitrary waveform reception*********/

// Enable (val == 1) or disable (val == 0) the capturing of digital waveform
// as timing data instead of DALI protocol decoding.
void DALISetReceiveTimingList(uint8_t val);

// This function returns 0xFF if no digital waveform capture has been completed.
// Otherwise (a waveform has been completed) it returns the location in the
// data buffer we're currently at (initially 0 and incremented after each
// DALIReadTimingListValue()).
uint8_t DALIReceiveTimingListDone(void);

// Read one value from the digital waveform timing list buffer. If it hits the
// end of the buffer, it will keep returning 0xFFFF to signal this.
uint16_t DALIReadTimingListValue(void);

/*********Arbitrary waveform reception**(END)**/


/*********DALI Data transmission and reception*********/

// Transmit command on the DALI bus. The machine will also wait for any reply
// after the transmission. The function returns 0 when successful and 1 to signal
// an error (machine busy).
uint8_t DALISendData(uint8_t addr, uint8_t val);

// Returns true if there's any data available in the input data buffer
uint8_t DALIDataAvailable(void);

// Function that returns the flags associated to the data being read. These
// give the status of this data transaction. This function moves the read
// pointer to the fresh data and returns the forward frame. It is required that
// out of the DALIReceiveDataFlags functions this be called first.
uint8_t DALIReceiveDataFlags(void);

// The data in the input data buffer is composed of forward data, backward data,
// timing data and flags. The user must first call DALIReceiveDataFlags.
uint16_t DALIReceiveDataForward(void);

// Function that returns the backward frame (if available) from the buffer.The validity of this value is
// reported in the flags.
uint8_t DALIReceiveDataBackward(void);

// Function that returns the delay between the forward and the backward  frames.
// The user must first call DALIReceiveDataFlags. The validity of this value
// is reported in the flags (it may be invalid if the data only contains a
// forward or a backward frame, but not both).
uint16_t DALIReceiveDataTiming(void);

/*********DALI Data transmission and reception**(END)**/



/*********DEBUG Functions*********/

// DEBUG: Function that returns the flags of the DALI protocol state machine
inline uint16_t DALIReadFlags(void);

// DEBUG: Function that returns the state of the DALI protocol state machine
inline uint8_t DALIReadState(void);

/*********DEBUG Functions**(END)**/



/**********************Public function definitions*(END)***********************/

#endif	/* DALI_H */

