/********************************************************************
 *
 * Copyright (C) 2014 Microchip Technology Inc. and its subsidiaries
 * (Microchip).  All rights reserved.
 *
 * You are permitted to use the software and its derivatives with Microchip
 * products. See the license agreement accompanying this software, if any, for
 * more info about your rights and obligations.
 *
 * SOFTWARE AND DOCUMENTATION ARE PROVIDED "AS IS" WITHOUT WARRANTY OF ANY KIND,
 * EITHER EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION, ANY WARRANTY OF
 * MERCHANTABILITY, TITLE, NON-INFRINGEMENT AND FITNESS FOR A PARTICULAR
 * PURPOSE. IN NO EVENT SHALL MICROCHIP, SMSC, OR ITS LICENSORS BE LIABLE OR
 * OBLIGATED UNDER CONTRACT, NEGLIGENCE, STRICT LIABILITY, CONTRIBUTION, BREACH
 * OF WARRANTY, OR OTHER LEGAL EQUITABLE THEORY FOR ANY DIRECT OR INDIRECT
 * DAMAGES OR EXPENSES INCLUDING BUT NOT LIMITED TO ANY INCIDENTAL, SPECIAL,
 * INDIRECT OR CONSEQUENTIAL DAMAGES, OR OTHER SIMILAR COSTS. To the fullest
 * extend allowed by law, Microchip and its licensors liability will not exceed
 * the amount of fees, if any, that you paid directly to Microchip to use this
 * software.
 *************************************************************************
 *
 *                           main.c
 *
 * About:
 *  This USB to DALI code is to demonstrate the DALI Library and to be a
 *  tool to assist with commissioning and testing of DALI fixtures.
 *  This firmware version may also be used for sniffing data.
 *
 *
 * Hardware:
 *  USB to DALI Interface - DM160215
 *
 *
 * Author            Date            Ver     Comment
 *~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
 * Michael Pearce    27 November 2012 0.00    Start of Code
 * Mihai Cuciuc      18 October 2013  1.00    Implemented DALI functionality
 *
 ******************************************************************************/

#include <xc.h>
#include <stdint.h>
#include "usb.h"
#include "usb_function_hid.h"
#include "dali.h"


// CONFIG1
#pragma config FOSC = INTOSC    // Oscillator Selection Bits (INTOSC oscillator: I/O function on CLKIN pin)
#pragma config WDTE = OFF       // Watchdog Timer Enable (WDT disabled)
#pragma config PWRTE = OFF      // Power-up Timer Enable (PWRT disabled)
#pragma config MCLRE = ON       // MCLR Pin Function Select (MCLR/VPP pin function is MCLR)
#pragma config CP = OFF         // Flash Program Memory Code Protection (Program memory code protection is disabled)
#pragma config BOREN = OFF      // Brown-out Reset Enable (Brown-out Reset disabled)
#pragma config CLKOUTEN = OFF   // Clock Out Enable (CLKOUT function is disabled. I/O or oscillator function on the CLKOUT pin)
#pragma config IESO = OFF       // Internal/External Switchover Mode (Internal/External Switchover Mode is disabled)
#pragma config FCMEN = OFF      // Fail-Safe Clock Monitor Enable (Fail-Safe Clock Monitor is disabled)

// CONFIG2
#pragma config WRT = OFF        // Flash Memory Self-Write Protection (Write protection off)
#pragma config CPUDIV = NOCLKDIV// CPU System Clock Selection Bit (NO CPU system divide)
#pragma config USBLSCLK = 48MHz // USB Low SPeed Clock Selection bit (System clock expects 48 MHz, FS/LS USB CLKENs divide-by is set to 8.)
#pragma config PLLMULT = 3x     // PLL Multipler Selection Bit (3x Output Frequency Selected)
#pragma config PLLEN = ENABLED  // PLL Enable Bit (3x or 4x PLL Enabled)
#pragma config STVREN = ON      // Stack Overflow/Underflow Reset Enable (Stack Overflow or Underflow will cause a Reset)
#pragma config BORV = LO        // Brown-out Reset Voltage Selection (Brown-out Reset Voltage (Vbor), low trip point selected.)
#pragma config LPBOR = OFF      // Low-Power Brown Out Reset (Low-Power BOR is disabled)
#pragma config LVP = OFF        // Low-Voltage Programming Enable (High-voltage on MCLR/VPP must be used for programming)

// Firmware version
#define FW_VER_MAJOR                    1
#define FW_VER_MINOR                    0


// USB LED registers
#define LED_USB                         LATC2
#define LED_USB_TRIS                    TRISC2

// DALI LED registers
#define LED_DALI                        LATC3
#define LED_DALI_TRIS                   TRISC3

#define LED_USB_COUNTER_LOAD            10      // Milliseconds that USB LED should stay lit after bus activity
#define LED_DALI_COUNTER_LOAD           10      // Milliseconds that DALI LED should stay lit after bus activity

// Locations of USB dual port RAM buffers
#define RX_DATA_BUFFER_ADDRESS          @0x2050
#define TX_DATA_BUFFER_ADDRESS          @0x20A0

// DALI flag masks.
#define DALI_COMMAND_RUNNING            0x01    // Command is being sent/received
#define DALI_COMMAND_DONE               0x02    // Command ready, PC may be told result
#define DALI_SEQUENCE_RUNNING           0x04    // Set of commands running
#define DALI_SEQUENCE_DONE              0x08    // Set of commands ready, PC may be told results
#define DALI_PING_DELAY                 0x10    // Delay expired, PC may be told so
#define DALI_ENABLE_DEBUG_PACKETS       0x20    // Debug packets are enabled, PC should be nagged

#define STATE_VECTOR_SIZE               10      // Length of buffer for state transitions



// Types of data packets that the PC can send to the device
enum
{
    TYPE_NOTHING,                       // Nothing to do
    TYPE_PING,                          // Ask for a reply in order to test the communication
    TYPE_DALI_COMMAND,                  // Send command on the DALI bus
    TYPE_STATE_QUERY,                   // [DEBUG] Ask the DALI state machine what state it's in
    TYPE_DALI_COMMAND_SEQUENCE,         // Send group of commands on the DALI bus
    TYPE_PING_DELAY,                    // Ask the device for a delayed reply (for accurate PC application timing)
    TYPE_SEND_DATA_TIMING_LIST,         // Send arbitrary waveform on the DALI bus
    TYPE_RESET_TIMING_LIST,             // Prepare to load arbitrary waveform data
    TYPE_LOAD_DATA_TIMING_LIST,         // Load arbitrary waveform data
    TYPE_SET_RECEIVE_TIMING_LIST,       // Enable/Disable waveform capture vs. decoded protocol
    TYPE_READ_FLAGS,                    // [DEBUG] Read DALI state machine flags
    TYPE_ENABLE_DEBUG_PACKETS,          // Enable/Disable the reception of debug data (state transitions)
    TYPE_GET_FW_VERSION                 // Request firmware version
};

// Tyeps of data packets that the device can send to the PC
enum
{
    REPLY_NOTHING,                      // Nothing done
    REPLY_PING,                         // Reply to ping request
    REPLY_DALI_ANSWER,                  // Confirmation of sent DALI command and DALI reply (if available)
    REPLY_STATE_QUERY,                  // [DEBUG] State that the DALI state machine is in
    REPLY_DALI_COMMAND_SEQUENCE,        // Confirmation of sent DALI group of commands and DALI replies (if available)
    REPLY_PING_DELAY,                   // Delayed reply
    REPLY_SEND_DATA_TIMING_LIST,        // Confirmation that arbitrary waveform is being sent on the DALI bus
    REPLY_RESET_TIMING_LIST,            // Confirmation of preparation to load arbitrary waveform data
    REPLY_LOAD_DATA_TIMING_LIST,        // Confirmation of loading arbitrary waveform data
    REPLY_SET_RECEIVE_TIMING_LIST,      // Confirmation of selection of waveform capture vs decoded protocol
    UNSOL_RECEIVE_TIMING_LIST,          // Captured waveform data from the DALI bus
    UNSOL_RECEIVE_DATA,                 // Captured decoded protocol data from the DALI bus
    REPLY_READ_FLAGS,                   // [DEBUG] DALI state machine flags
    UNSOL_STATE_CHANGE,                 // [DEBUG] State machine transition or flag being toggled
    REPLY_ENABLE_DEBUG_PACKETS,         // Confirmation of selection of debug data
    REPLY_GET_FW_VERSION                // Reply with firmware version
};

// Counters for how long each LED should be lit, in ms. These can be loaded by
// the application to light the LED for that period.
uint8_t ledUsbCounter;
uint8_t ledDaliCounter;

// USB dual port RAM buffers, which are located at fixed addresses in the CPU
// linear address space
uint8_t usbReceivedDataBuffer[64] RX_DATA_BUFFER_ADDRESS;
uint8_t usbToSendDataBuffer[64] TX_DATA_BUFFER_ADDRESS;

//USB handles.  Must be initialized to 0 at startup.
USB_HANDLE USBOutHandle = 0;
USB_HANDLE USBInHandle = 0;

// DALI flags
volatile uint8_t daliFlags;

// The DALI command sequence machinery uses these variables for commands and
// answers storage and position in the buffers
uint8_t daliCommandNumber, daliCommandCount;
uint8_t daliCommandBuffer[60];
uint8_t daliAnswerBuffer[60];

// The PC may request a delayed ping to respect strict timing. This variable
// keeps the number of milliseconds to wait until sending this packet
volatile uint16_t pingDelay;

// When debug data is enabled, the FW keeps all DALI state/flag transitions in
// these buffers and sends them to the PC application as soon as possible. The
// R and W variables keep the reading and writing positions in the circular buffers
volatile uint8_t stateVector[STATE_VECTOR_SIZE];
volatile uint16_t flagsVector[STATE_VECTOR_SIZE];
volatile uint8_t stateVectorR, stateVectorW;

// These variables keep the previous DALI state/flags in order to be able to
// detect changes
uint8_t oldState;
uint16_t oldFlags;

// The DALI protocol state machine generates this type of data along with the
// actual DALI bus data. This is what tells the application what type of data
// has been received and if it's valid.
bundle_flags_type flags;


// Hardware initialization
void InitializeSystem(void);

// Processing of USB data. This needs to be called from while(1)
void ProcessUSB(void);

// Initialization of Timer2 to generate 1ms interrupts
void TickInit(void);



void main(void)
{
    // Initialize application hardware
    InitializeSystem();

    // Initialize Timer2 to provide 1ms interrupts
    TickInit();

    // The DALI machinery should not do anything on start-up
    daliFlags = 0;

    // Enable interrupts
    PEIE=1;
    GIE=1;

    // USB specific initialization
    USBDeviceAttach();

    // Initialize DALI hardware (timer and external interrupt) and  protocol
    // state machine
    DALIInit();

    // Debug data buffer initialization
    stateVectorR = 0;
    stateVectorW = 0;
    oldState = 0xFF;
    oldFlags = 0xFFFF;

    while(1)
    {
        if ((daliFlags & DALI_ENABLE_DEBUG_PACKETS) != 0)
        {
            // If DEBUG information is enabled by the PC application, we need
            // to keep polling the DALI protocol state machine and its flags
            if ((DALIReadState() != oldState) || (DALIReadFlags() != oldFlags))
            {
                // A change has been found, we append it to the circular buffer
                // if we have enough room for it (do not overwrite old data if
                // it hasn't been sent to the PC yet)
                oldState = DALIReadState();
                oldFlags = DALIReadFlags();
                if ((stateVectorW + 1) % STATE_VECTOR_SIZE != stateVectorR)
                {
                    stateVectorW = (stateVectorW + 1) % STATE_VECTOR_SIZE;
                    stateVector[stateVectorW] = oldState;
                    flagsVector[stateVectorW] = oldFlags;
                }
            }
        }

        // This function checks the USB handles and if data has been received
        // acts on it. Also, it checks the application flags for any data that
        // needs to be sent to the PC and when this can be done is initiated.
        // As such, this needs to be called as often as possible from the
        // mainline code
        ProcessUSB();

        // If we've started sending a command over the DALI bus we need to see
        // if it's ready
        if ((daliFlags & DALI_COMMAND_RUNNING) != 0)
        {
            // Since transmitting data over the DALI bus takes a long time (~15ms
            // just the forward frame) we keep polling the DALI protocol state
            // machine to find out when it finishes sending data.
            if (DALIReadState() == ST_IDLE)
            {
                // The protocol has finished sending the data and receiving any
                // reply that might have come. We now need to test if this is
                // part of a series of commands that need to be sent quickly one
                // after the other, or it's just a single command.
                if ((daliFlags & DALI_SEQUENCE_RUNNING) == 0)
                {
                    // No sequence is running. Set COMMAND_DONE flag to allow the
                    // USB to send the reply to the PC
                    daliFlags &= ~DALI_COMMAND_RUNNING;
                    daliFlags |= DALI_COMMAND_DONE;
                }
                else
                {
                    // Command was part of a sequence. Check result, then advance
                    // or set the output accordingly to allow the USB to send
                    // the results
                    if (DALIDataAvailable())
                    {
                        daliFlags &= ~DALI_COMMAND_RUNNING;
                        
                        // Read flags and advance pointer
                        flags.flagsByte = DALIReceiveDataFlags();
                        // Check if this command received a response and did not
                        // generate any error
                        if (flags.txThisDevice & flags.backwardFrameValid & (~flags.txError) & (~flags.rxError) & (~flags.rxTimingError))
                        {
                            // Answer received. This flag is placed single on a
                            // byte for easy verification in the PC UI
                            daliAnswerBuffer[daliCommandNumber * 3] = 1;
                        }
                        else
                        {
                            // No answer received
                            daliAnswerBuffer[daliCommandNumber * 3] = 0;
                        }
                        // The PC will also get a copy of all flags...
                        daliAnswerBuffer[(daliCommandNumber * 3) + 1] = flags.flagsByte;
                        // ... and the reply itself (which may or may not be
                        // valid, based on the flags)
                        daliAnswerBuffer[(daliCommandNumber * 3) + 2] = DALIReceiveDataBackward();

                        // Advance in the command buffer
                        daliCommandNumber++;
                        // Did we finish?
                        if (daliCommandNumber < daliCommandCount)
                        {
                            // More commands. Retrigger command on the DALI bus
                            DALISendData(daliCommandBuffer[(daliCommandNumber * 3) + 1], daliCommandBuffer[(daliCommandNumber * 3) + 2]);
                            // Tell ourselves that we have a command running
                            daliFlags |= DALI_COMMAND_RUNNING;
                            // And light the DALI LED as well
                            ledDaliCounter = LED_DALI_COUNTER_LOAD;
                        }
                        else
                        {
                            // This was the last command.
                            daliFlags &= ~DALI_SEQUENCE_RUNNING;
                            daliFlags |= DALI_SEQUENCE_DONE;
                        }
                    }
                }
            }
        }
    }
}


// Initialization of Timer2 to generate 1ms interrupts
void TickInit(void)
{
    // Use Timer 2 to generate 1ms Interrupt tick
    // 48MHz/4(inst)/16(pre) = 750kHz, 750KHz / 10(post) = 75; PR2 = 74 => 1KHz (1ms)
    PR2 = 74;
    TMR2 = 0;
    T2CON = 0b01001110; // Post=1:10, On, Pre=16.
    TMR2IF = 0;
    TMR2IE = 1;
}

// Hardware initialization
void InitializeSystem(void)
{
    ANSELA = 0x00;
    ANSELC = 0x00;
    TRISA  = 0xFF;  // All inputs until set individually
    TRISC  = 0xFF;

    #ifdef _16F1459
    ANSELB = 0x00;
    TRISB  = 0xFF;
    #endif

    LED_USB = 0;
    LED_USB_TRIS = 0;
    LED_DALI = 0;
    LED_DALI_TRIS = 0;
    ledUsbCounter = 0;
    ledDaliCounter = 0;

    OSCTUNE = 0;
    OSCCON = 0x7C;   // PLL enabled, 3x, 16MHz internal osc, SCS external
    OSCCONbits.SPLLMULT = 1;   // 1=3x, 0=4x
    ACTCON = 0x90;   // Clock recovery on, Clock Recovery enabled; SOF packet

    USBOutHandle = 0;
    USBInHandle = 0;

    USBDeviceInit();	//usb_device.c.  Initializes USB module SFRs and firmware
    					//variables to known states.
}


// Processing of USB data. This needs to be called from while(1)
void ProcessUSB(void)
{
    static char txPacket = 0;
    uint8_t i;
    uint16_t val;

    // If the device hasn't yet been fully enumerated don't do anything
    if((USBDeviceState < CONFIGURED_STATE)||(USBSuspendControl==1))
    {
        return;
    }


    if(!HIDTxHandleBusy(USBInHandle) && (txPacket == 0))
    {
        // Here we're sure we're not transmitting anything and the application
        // hasn't yet prepared anything for transmission either.


        if (!USBHandleBusy(USBOutHandle))
        {
            usbToSendDataBuffer[0] = 0;
            usbToSendDataBuffer[1] = 0;
            usbToSendDataBuffer[2] = 0;
        }

        if (((daliFlags & DALI_PING_DELAY) != 0) && (pingDelay == 0))
        {
            // The PC application has requested a delayed ping and the timer
            // expired, thus we need to send a reply.
            daliFlags &= ~DALI_PING_DELAY;
            usbToSendDataBuffer[0] = REPLY_PING_DELAY;
            txPacket = 1;
        }
        else if ((daliFlags & DALI_COMMAND_DONE) != 0)
        {
            if (DALIDataAvailable())
            {
                // The DALI protocol state machine is ready after sending a command
                // and it also filled in some data in the buffer. This need not
                // be a backward frame sent by a Control Gear -- any data sent
                // on the DALI bus triggers some data reply which contains the
                // status of the transmission.
                daliFlags &= ~DALI_COMMAND_DONE;
                usbToSendDataBuffer[0] = REPLY_DALI_ANSWER;

                // Read the flags. The DALIReceiveDataFlags() function also
                // moves the internal pointer to the new data, such that the
                // other DALIReceiveData functions get fresh data.
                flags.flagsByte = DALIReceiveDataFlags();
                
                // From the flags content determine if the command triggered a
                // reply
                if (flags.txThisDevice & flags.backwardFrameValid & (~flags.txError) & (~flags.rxError) & (~flags.rxTimingError))
                {
                    usbToSendDataBuffer[1] = 1;
                }
                else
                {
                    usbToSendDataBuffer[1] = 0;
                }

                // Format the rest of the data for transmission to the PC
                usbToSendDataBuffer[2] = DALIReceiveDataBackward();
                usbToSendDataBuffer[3] = flags.flagsByte;

                // Load the forward data (what we've sent)
                val = DALIReceiveDataForward();
                usbToSendDataBuffer[4] = val >> 8;
                usbToSendDataBuffer[5] = val & 0xFF;

                val = DALIReceiveDataTiming();
                usbToSendDataBuffer[6] = val >> 8;
                usbToSendDataBuffer[7] = val & 0xFF;
                // Signal that a packet will need to be sent
                txPacket = 1;
            }
        }
        else if ((daliFlags & DALI_SEQUENCE_DONE) != 0)
        {
            // Our machinery has finished sending a sequence of commands. Now
            // a sequence of flags and replies (if available) are in the
            // daliAnswerBuffer vector and these need to be sent to the PC for
            // analysis
            daliFlags &= ~DALI_SEQUENCE_DONE;
            usbToSendDataBuffer[0] = REPLY_DALI_COMMAND_SEQUENCE;
            usbToSendDataBuffer[1] = daliCommandCount;
            for (daliCommandNumber = 0; daliCommandNumber < (daliCommandCount * 3); daliCommandNumber++)
            {
                usbToSendDataBuffer[daliCommandNumber + 2] = daliAnswerBuffer[daliCommandNumber];
            }
            txPacket = 1;
        }
        else if (DALIReceiveTimingListDone() != 0xFF)
        {
            // The FW has been configured for digital waveform capture, instead
            // of protocol decoding. It captured such a waveform and wants to
            // send it to the PC application. We do this now.
            usbToSendDataBuffer[0] = UNSOL_RECEIVE_TIMING_LIST;
            // The DALIReceiveTimingListDone() function returns the index of the
            // 16-bit values captured we're currently at. This makes it easy
            // for the PC UI to append multiple such messages into the same
            // array.
            usbToSendDataBuffer[1] = DALIReceiveTimingListDone();
            i = 2;
            // Add 30 such values to the buffer and send them to the PC
            while (i < 60)
            {
                val = DALIReadTimingListValue();
                usbToSendDataBuffer[i] = val >> 8;
                usbToSendDataBuffer[i + 1] = val & 0xFF;
                i += 2;
            }
            txPacket = 1;
        }
        else if (((daliFlags & DALI_COMMAND_RUNNING) == 0) && (DALIDataAvailable() != 0))
        {
            // Unsolicited data on the bus (i.e. transaction that we did not
            // initiate). Add everything about it to the buffer and send it to
            // the PC
            usbToSendDataBuffer[0] = UNSOL_RECEIVE_DATA;

            // Start with the flags such that the following functions get fresh
            // data.
            usbToSendDataBuffer[4] = DALIReceiveDataFlags();

            val = DALIReceiveDataForward();
            usbToSendDataBuffer[1] = val >> 8;
            usbToSendDataBuffer[2] = val & 0xFF;

            usbToSendDataBuffer[3] = DALIReceiveDataBackward();
            val = DALIReceiveDataTiming();

            usbToSendDataBuffer[5] = val >> 8;
            usbToSendDataBuffer[6] = val & 0xFF;
            txPacket = 1;
        }
        else if (stateVectorW != stateVectorR)
        {
            // If the vector of states/flags is not empty (i.e. we have enabled
            // DEBUG packets and some have piled up) send one frame to the PC
            stateVectorR = (stateVectorR + 1) % STATE_VECTOR_SIZE;
            usbToSendDataBuffer[0] = UNSOL_STATE_CHANGE;
            usbToSendDataBuffer[1] = stateVector[stateVectorR];
            usbToSendDataBuffer[2] = stateVectorW;
            usbToSendDataBuffer[3] = stateVectorR;
            usbToSendDataBuffer[4] = 0;
            usbToSendDataBuffer[5] = 0;
            usbToSendDataBuffer[6] = (flagsVector[stateVectorR] >> 8) & 0xFF;
            usbToSendDataBuffer[7] = flagsVector[stateVectorR] & 0xFF;
            txPacket = 1;
        }


        // Check if we have received an OUT data packet from the host
        if (!HIDRxHandleBusy(USBOutHandle))
        {
            // We just received a packet of data from the USB host.
            // Check the first byte of the packet to see what command the host
            // application software wants us to fulfill.

            switch(usbReceivedDataBuffer[0])
            {

                case TYPE_NOTHING:
                    // Nothing to do
                    break;

                case TYPE_GET_FW_VERSION:
                    // We should reply with our firmware version
                    usbToSendDataBuffer[0] = REPLY_GET_FW_VERSION;
                    usbToSendDataBuffer[1] = FW_VER_MAJOR;
                    usbToSendDataBuffer[2] = FW_VER_MINOR;
                    txPacket = 1;
                    break;

                case TYPE_PING:
                    // Ping request. Just reply with something
                    usbToSendDataBuffer[0] = REPLY_PING;
                    usbToSendDataBuffer[1] = 0xAA;
                    txPacket = 1;
                    break;

                case TYPE_ENABLE_DEBUG_PACKETS:
                    // Enable/Disable the reception of debug data (state
                    // transitions). Also send a confirmation to the PC with the
                    // modified value
                    usbToSendDataBuffer[0] = REPLY_ENABLE_DEBUG_PACKETS;
                    if (usbReceivedDataBuffer[1] == 0)
                    {
                        daliFlags &= ~DALI_ENABLE_DEBUG_PACKETS;
                        usbToSendDataBuffer[1] = 0;
                    }
                    else
                    {
                        daliFlags |= DALI_ENABLE_DEBUG_PACKETS;
                        usbToSendDataBuffer[1] = 1;
                    }
                    txPacket = 1;
                    break;

                case TYPE_SEND_DATA_TIMING_LIST:
                    // Trigger the sending of the previously loaded timing list
                    DALISendDataTimingList();
                    ledDaliCounter = LED_DALI_COUNTER_LOAD;
                    usbToSendDataBuffer[0] = REPLY_SEND_DATA_TIMING_LIST;
                    txPacket = 1;
                    break;

                case TYPE_RESET_TIMING_LIST:
                    // Reset timing list buffer position, such that we can start
                    // loading new data from the beginning of the list
                    DALIResetTimingList();
                    usbToSendDataBuffer[0] = REPLY_RESET_TIMING_LIST;
                    txPacket = 1;
                    break;

                case TYPE_READ_FLAGS:
                    // Read the DALI protocol flags, for debug purposes
                    usbToSendDataBuffer[0] = REPLY_READ_FLAGS;
                    usbToSendDataBuffer[1] = 0;
                    usbToSendDataBuffer[2] = 0;
                    usbToSendDataBuffer[3] = (DALIReadFlags() >> 8) & 0xFF;
                    usbToSendDataBuffer[4] = DALIReadFlags() & 0xFF;
                    txPacket = 1;
                    break;

                case TYPE_LOAD_DATA_TIMING_LIST:
                    // Load data into the timing list buffer, in order to send
                    // a digital waveform on the DALI bus at a later time
                    i = 2;
                    while (i < 60)
                    {
                        val = (usbReceivedDataBuffer[i] << 8) | usbReceivedDataBuffer[i+1];
                        DALILoadTimingListValue(val);
                        if (val == 0xFFFF)
                        {
                            break;
                        }
                        i += 2;
                    }
                    usbToSendDataBuffer[0] = REPLY_LOAD_DATA_TIMING_LIST;
                    txPacket = 1;
                    break;

                case TYPE_SET_RECEIVE_TIMING_LIST:
                    // Enable/disable the reception of timing list values (digital
                    // waveform) instead of the decoded DALI protocol.
                    DALISetReceiveTimingList(usbReceivedDataBuffer[1]);
                    usbToSendDataBuffer[0] = REPLY_SET_RECEIVE_TIMING_LIST;
                    usbToSendDataBuffer[1] = usbReceivedDataBuffer[1];
                    txPacket = 1;
                    break;

                case TYPE_PING_DELAY:
                    // The PC asked us to delay for a number of milliseconds
                    // (max 65535) and give it a nudge once the delay expired
                    pingDelay = (usbReceivedDataBuffer[1] << 8) | usbReceivedDataBuffer[2];
                    daliFlags |= DALI_PING_DELAY;
                    break;

                case TYPE_DALI_COMMAND:
                    // Send one DALI command on the DALI bus. This may trigger
                    // a reply which will be handled once the command is finished
                    DALISendData(usbReceivedDataBuffer[1], usbReceivedDataBuffer[2]);
                    daliFlags |= DALI_COMMAND_RUNNING;
                    ledDaliCounter = LED_DALI_COUNTER_LOAD;
                    break;

                case TYPE_STATE_QUERY:
                    // Ask the DALI protocol state machine what state it's it,
                    // for debug purposes
                    usbToSendDataBuffer[0] = REPLY_STATE_QUERY;
                    usbToSendDataBuffer[1] = DALIReadState();
                    txPacket = 1;
                    break;

                case TYPE_DALI_COMMAND_SEQUENCE:
                    // Send a sequence of DALI commands tightly packed (as soon
                    // as the machine is ready sending one command, send the next)
                    daliCommandCount = usbReceivedDataBuffer[1];
                    for (daliCommandNumber = 0; daliCommandNumber < (daliCommandCount * 3); daliCommandNumber++)
                    {
                        daliCommandBuffer[daliCommandNumber] = usbReceivedDataBuffer[daliCommandNumber + 2];
                    }
                    daliCommandNumber = 0;
                    DALISendData(daliCommandBuffer[1], daliCommandBuffer[2]);
                    daliFlags |= DALI_SEQUENCE_RUNNING | DALI_COMMAND_RUNNING;
                    ledDaliCounter = LED_DALI_COUNTER_LOAD;
                    break;
            }

            // Re-arm the OUT endpoint, so we can receive the next OUT data packet
            // that the host may try to send us.
            USBOutHandle = HIDRxPacket(HID_EP, (BYTE*)&usbReceivedDataBuffer, 64);
            ledUsbCounter = LED_USB_COUNTER_LOAD;
        }
    }

    // If we need to send something and we may do it, tell the USB stack we're
    // ready
    if(!HIDTxHandleBusy(USBInHandle) && (txPacket == 1))
    {
        USBInHandle = HIDTxPacket(HID_EP,(BYTE*)&usbToSendDataBuffer[0],64);
        txPacket = 0;
        ledUsbCounter = LED_USB_COUNTER_LOAD;
    }

}



void interrupt ISRCode()
{
    // USB stack ISR code
    USBDeviceTasks();

    // Application 1ms tick. The DALI Stack also checks this flag and will clear it.
    if (TMR2IF != 0)
    {
        // Handle the USB led
        if (ledUsbCounter == 0)
        {
            LED_USB = 0;
        }
        else
        {
            LED_USB = 1;
            ledUsbCounter--;
        }

        // Handle the DALI led
        if (ledDaliCounter == 0)
        {
            LED_DALI = 0;
        }
        else
        {
            LED_DALI = 1;
            ledDaliCounter--;
        }

        // Timer for the PC application
        if(pingDelay != 0)
        {
            pingDelay--;
        }
    }

    // DALI protocol interrupt code. This handles the external interrupt (DALI
    // bus transition) and timer interrupt (time-outs), as well as detection
    // of cable disconnection
    DALIInterrupt();
}
